"""Python version of object feature extractor, handles extra features.
   jiefeng @ 2016-02-19
"""

import os
import sys
import commentjson
import cv2
from common import *
from cnn_feats import *

cur_dir = os.path.dirname(os.path.abspath(__file__)) + '/'
engine_root = os.path.join(cur_dir, '../../../')
engine_root = 'E:/Projects/Github/Owl/Owl/'
print engine_root

engine_cpp_lib_path = os.path.join(engine_root, 'CoreEngine/EngineLib/lib/')
sys.path.append(engine_cpp_lib_path)
from EngineLib import ObjFeatExtractorCPP

engine_python_lib_path = os.path.join(engine_root,
                                      'CoreEngine/EngineLibPython/')
sys.path.append(engine_python_lib_path)
from database.DBManager import MongoManager


class UniversalObjFeatExtractor(object):
  """Python interface for feature extraction.

    Universal for c++ and python, should be used for database usage
    usage: init from config file, creat corresponding params and feature extractor
    """

  obj_feat_extractor_cpp = ObjFeatExtractorCPP()
  feat_params = []
  feat_computers = []
  raw_img_bgr = None
  inside_img = None
  config_fn = ''
  extract_cpp = True
  extract_python = True

  # load feature info from json config file
  def Init(self, config_fn):
    self.config_fn = config_fn

    # init cpp part
    self.obj_feat_extractor_cpp.init(config_fn)

    # init python part
    with open(config_fn, 'r') as f:
      config = commentjson.load(f)

    # extract features
    feat_configs = config['engine']['features']
    for feat_config in feat_configs:
      if not feat_config['enabled']:
        continue

      if IsCNNFeat(feat_config['type']):
        # create params
        cur_params = CNNParams()
        cur_params.feat_name = feat_config['type'].encode("utf-8")
        print 'using feature: {}'.format(cur_params.feat_name)
        cur_params.feat_type = MapNameToFeatType(cur_params.feat_name)
        self.feat_params.append(cur_params)
        # create feature
        cnn_feat_computer = DeepCNNFeature()
        cnn_feat_computer.Init(cur_params)
        cnn_feat_computer.Start()
        self.feat_computers.append(cnn_feat_computer)

    print 'feature config loaded'

  def Start(self):
    pass

  # def SetInput(self, img_fn):
  #     self.obj_feat_extractor_cpp.set_input(img_fn)
  #     self.raw_img_bgr = cv2.imread(img_fn)

  # return a dictionary of named features
  def Compute(self, img_fn, mask_fn):
    feats = {}
    # cpp
    if self.extract_cpp:
      feats = self.obj_feat_extractor_cpp.compute(
          img_fn.encode('utf-8'), mask_fn.encode('utf-8'))
    # python
    if self.extract_python:
      for i in range(len(self.feat_params)):
        feat_name = self.feat_params[i].feat_name
        feats[feat_name] = self.feat_computers[i].ComputeForFns(
            [img_fn], [mask_fn]).tolist()[0]

    return feats

  def PopulateDB(self):
    print 'start populating database'

    # connect db
    db_man = MongoManager()
    db_man.load_config(self.config_fn)
    db_man.connect()
    print 'database connected'

    # compute for each item
    total_cnt = db_man.collection.count()
    print 'total object to process: {}'.format(total_cnt)
    valid_num = 0
    cursor = db_man.collection.find()
    for item in cursor:
      try:
        img_fn = item['img_file_path']
        mask_fn = item['mask_file_path']
        cur_feats = self.Compute(img_fn, mask_fn)
        # update each feature
        for feat_name, feat_val in cur_feats.iteritems():
          db_man.update({
              '_id': item['_id']
          }, 'visual_features.' + feat_name, feat_val)
        valid_num += 1
        if valid_num % 1000 == 0:
          print 'finished {}/{}'.format(valid_num, total_cnt)
      except Exception as ex:
        print 'error in {}: {}'.format(img_fn, ex.message)
        continue

    print 'feature extraction done. {}/{} finished. Percentage: {}%'.format(
        valid_num, total_cnt, valid_num * 100.0 / total_cnt)


if __name__ == '__main__':
  config_fn = os.path.join(engine_root,
                           'settings/engine_settings_2016_03.json')
  obj_feat_extractor = UniversalObjFeatExtractor()
  obj_feat_extractor.Init(config_fn)
  obj_feat_extractor.PopulateDB()



"""Binarization for saliency map.
"""

import numpy as np


def binarise_saliency_map(saliency_map, method='adaptive', threshold=0.5):

  # check if input is a numpy array
  if type(saliency_map).__module__ != np.__name__:
    print 'Expected numpy array'
    return None

  #check if input is 2D
  if len(saliency_map.shape) != 2:
    print('Saliency map must be 2D')
    return None

  if method == 'fixed':
    return (saliency_map > threshold)

  elif method == 'adaptive':
    adaptive_threshold = 2.0 * saliency_map.mean()
    return (saliency_map > adaptive_threshold)

  elif method == 'clustering':
    print('Not yet implemented')
    return None

  else:
    print("Method must be one of fixed, adaptive or clustering")
    return None



"""Shared definition for matcher.
"""

import abc


class MatcherType(object):
  """Built-in matcher type.
  """
  L2 = 0
  L1 = 1
  HAMMING = 2
  ANGULAR = 3


class MatchProtocol(object):
  """How to conduct match.
  """
  EXACT = 0
  APPROXIMATE = 1


class MatcherParams(object):
  """Parameters for matcher.
  """
  matcher_type = MatcherType.L2
  matcher_protocol = MatchProtocol.APPROXIMATE


class BaseMatcher(object):
  """Base class for a matcher.

  A matcher computes matches a set of queries to a database and
  return a ranked list of the item.
  """
  __metaclass__ = abc.ABCMeta

  matcher_type = None
  matcher_name = ""
  match_protocol = MatchProtocol.EXACT
  res_dir = None

  @abc.abstractmethod
  def __init__(self, name, matcher_type, res_dir=None):
    self.matcher_name = name
    self.res_dir = res_dir
    self.matcher_type = matcher_type

  @abc.abstractmethod
  def prepare_database(self, db_items):
    """Load database items.

    Args:
      db_items: database items, can take any form based
      on custom implementation.
    """
    pass

  @abc.abstractmethod
  def match(self, query_items, top_nn=10, dist_func=None):
    """Perform match on the database for each query.

    Args:
      query_items: a set of query image data. Can take any form.
      top_nn: top results to return.
      dist_func: custom distance function.
    Returns:
      ranked list of db items for each query.
    """
    pass



"""Evaluation pipeline for feature matching.
"""


class MatcherEvaluator(object):
  """Class to manage evaluation of a matcher.
  """
  matcher = None

  def __init__(self):
    pass

  def vis_matching(self):
    """Visualize matching result in a webpage.
    """
    pass

  def run_eval_pipeline_local(self, config_fn):
    """Local version of the evaluation pipeline.

    It takes in database and query data and produce evaluation
    results: a data file with eval summary and a webpage showing
    matching results.

    Args:
      config_fn: evaluation config file.
    See sample_config_fn.json for an example.
    """
    pass




"""Test saliency methods.
"""

import rbd
import mbd
import ft
import binarise as bina
from PIL import Image
import time
import skimage.io

# path to the image
filename = 'bird.jpg'
img_arr = skimage.io.imread(filename)

startt = time.time()
# get the saliency maps using the 3 implemented methods
rbd_res = rbd.get_saliency_rbd(img_arr).astype('uint8')
print "rbd time: {}s".format(time.time() - startt)

startt = time.time()
ft_res = ft.get_saliency_ft(img_arr).astype('uint8')
print "ft time: {}s".format(time.time() - startt)

startt = time.time()
mbd_res = mbd.get_saliency_mbd(img_arr).astype('uint8')
print "mbd time: {}s".format(time.time() - startt)

# often, it is desirable to have a binary saliency map
binary_sal = bina.binarise_saliency_map(mbd_res, method='adaptive')

img = Image.open(filename)
# img = cv2.imread(filename)

img.show(title="input")
rbd_res = Image.fromarray(rbd_res)
ft_res = Image.fromarray(ft_res)
mbd_res = Image.fromarray(mbd_res)
rbd_res.show("rbd")
ft_res.show("ft")
mbd_res.show("mbd")

#openCV cannot display numpy type 0, so convert to uint8 and scale
binary = Image.fromarray(255 * binary_sal.astype('uint8'))
binary.show("binary")



"""Test tumblr api.
"""

import unittest

from owl.third.social import tumblr


class TumblrAPITester(unittest.TestCase):
  def test_client(self):
    api = tumblr.TumblrAPI()
    items = api.crawl_posts("street style", to_continue=True)
    items2 = api.crawl_posts("street style", to_continue=True)
    items.extend(items2)
    for item in items:
      print item["img_urls"]
      print ""


if __name__ == "__main__":
  unittest.main()



"""OOP interface for S3.
"""

from cStringIO import StringIO

from owl.third.aws import common as aws_common
from owl.data import img_tools


class S3IO(object):
  """Class for managing s3 tasks.
  """
  session = None
  resource = None
  client = None
  bucket = None

  def __init__(self, session=None, access_key="", secret_key="", region=""):
    """Create s3 client.
    """
    if session is not None:
      self.session = session
    else:
      self.session = aws_common.init_aws(access_key, secret_key, region)
    self.client = self.session.client("s3")
    self.resource = self.session.resource("s3")

  def use_bucket(self, bucket_name):
    """Use a specified bucket.

    If the bucket doesn't exist, create one.
    """
    self.bucket = self.resource.create_bucket(Bucket=bucket_name)

  def upload_img_bin_as_file(self,
                             bucket_name,
                             img_bin,
                             img_format="png",
                             make_public=True):
    """Upload image binary data to bucket.

    Args:
      bucket_name: name of bucket.
      img_bin: image binary data.
      img_format: type of image, only support jpy or png.
      make_public: whether to make it publically accessible.
    Returns:
      url of the image on s3.
    """
    assert img_format in ["png", "jpg", "jpeg"
                          ], "img_format must be either png, jpg or jpeg."
    img_sha1 = img_tools.img_bin_to_sha1(img_bin)
    img_key_name = "{}.{}".format(img_sha1, img_format)
    upload_args = {}
    if img_format in ["jpg", "jpeg"]:
      upload_args["ContentType"] = "image/jpeg"
    if img_format == "png":
      upload_args["ContentType"] = "image/png"
    if make_public:
      upload_args["ACL"] = "public-read"
    self.bucket.upload_fileobj(
        StringIO(img_bin), img_key_name, ExtraArgs=upload_args)
    img_url = "{}/{}/{}".format(self.client.meta.endpoint_url, bucket_name,
                                img_key_name)
    return img_url

  def save_img(self, bucket_name, img_base64):
    """Save image data to bucket.

    sha1 is used as key for image data.

    Args:
      bucket_name: name of the target bucket.
      img_base64: base64 string of image.
    Returns:
      image sha1 hash as key.
    """
    img_sha1 = img_tools.base64_to_sha1(img_base64)
    self.bucket.put_object(Key=img_sha1, Body=img_base64)
    return img_sha1

  def get_object(self, bucket_name, obj_key):
    """Get object binary data from bucket.

    Args:
      bucket_name: name of the target bucket.
      obj_key: object key.
    Returns:
      object data.
    """
    res = self.client.get_object(Bucket=bucket_name, Key=obj_key)
    obj_data = res["Body"].read()
    return obj_data



"""Implementation of "Minimum Barrier Salient Object Detection at 80 FPS",
by: Jianming Zhang, Stan Sclaroff, Zhe Lin, Xiaohui Shen, Brian Price and Radomir Mech.
Code from https://github.com/yhenon/pyimgsaliency.
"""

import math
import sys
import operator
import networkx as nx
import numpy as np
import scipy.spatial.distance
import scipy.signal
import skimage
import skimage.io
from skimage.segmentation import slic
from skimage.util import img_as_float


def raster_scan(img, L, U, D):
  n_rows = len(img)
  n_cols = len(img[0])

  for x in xrange(1, n_rows - 1):
    for y in xrange(1, n_cols - 1):
      ix = img[x][y]
      d = D[x][y]

      u1 = U[x - 1][y]
      l1 = L[x - 1][y]

      u2 = U[x][y - 1]
      l2 = L[x][y - 1]

      b1 = max(u1, ix) - min(l1, ix)
      b2 = max(u2, ix) - min(l2, ix)

      if d <= b1 and d <= b2:
        continue
      elif b1 < d and b1 <= b2:
        D[x][y] = b1
        U[x][y] = max(u1, ix)
        L[x][y] = min(l1, ix)
      else:
        D[x][y] = b2
        U[x][y] = max(u2, ix)
        L[x][y] = min(l2, ix)

  return True


def raster_scan_inv(img, L, U, D):
  n_rows = len(img)
  n_cols = len(img[0])

  for x in xrange(n_rows - 2, 1, -1):
    for y in xrange(n_cols - 2, 1, -1):

      ix = img[x][y]
      d = D[x][y]

      u1 = U[x + 1][y]
      l1 = L[x + 1][y]

      u2 = U[x][y + 1]
      l2 = L[x][y + 1]

      b1 = max(u1, ix) - min(l1, ix)
      b2 = max(u2, ix) - min(l2, ix)

      if d <= b1 and d <= b2:
        continue
      elif b1 < d and b1 <= b2:
        D[x][y] = b1
        U[x][y] = max(u1, ix)
        L[x][y] = min(l1, ix)
      else:
        D[x][y] = b2
        U[x][y] = max(u2, ix)
        L[x][y] = min(l2, ix)

  return True


def mbd(img, num_iters):

  if len(img.shape) != 2:
    print('did not get 2d np array to fast mbd')
    return None
  if (img.shape[0] <= 3 or img.shape[1] <= 3):
    print('image is too small')
    return None

  L = np.copy(img)
  U = np.copy(img)
  D = float('Inf') * np.ones(img.shape)
  D[0, :] = 0
  D[-1, :] = 0
  D[:, 0] = 0
  D[:, -1] = 0

  # unfortunately, iterating over numpy arrays is very slow
  img_list = img.tolist()
  L_list = L.tolist()
  U_list = U.tolist()
  D_list = D.tolist()

  for x in xrange(0, num_iters):
    if x % 2 == 1:
      raster_scan(img_list, L_list, U_list, D_list)
    else:
      raster_scan_inv(img_list, L_list, U_list, D_list)

  return np.array(D_list)


def get_saliency_mbd(img_arr, method='b'):
  img = img_arr
  img_mean = np.mean(img, axis=(2))
  sal = mbd(img_mean, 3)

  if method == 'b':
    # get the background map
    # paper uses 30px for an image of size 300px, so we use 10%
    (n_rows, n_cols, n_channels) = img.shape
    img_size = math.sqrt(n_rows * n_cols)
    border_thickness = int(math.floor(0.1 * img_size))

    img_lab = img_as_float(skimage.color.rgb2lab(img))

    px_left = img_lab[0:border_thickness, :, :]
    px_right = img_lab[n_rows - border_thickness - 1:-1, :, :]

    px_top = img_lab[:, 0:border_thickness, :]
    px_bottom = img_lab[:, n_cols - border_thickness - 1:-1, :]

    px_mean_left = np.mean(px_left, axis=(0, 1))
    px_mean_right = np.mean(px_right, axis=(0, 1))
    px_mean_top = np.mean(px_top, axis=(0, 1))
    px_mean_bottom = np.mean(px_bottom, axis=(0, 1))

    px_left = px_left.reshape((n_cols * border_thickness, 3))
    px_right = px_right.reshape((n_cols * border_thickness, 3))

    px_top = px_top.reshape((n_rows * border_thickness, 3))
    px_bottom = px_bottom.reshape((n_rows * border_thickness, 3))

    cov_left = np.cov(px_left.T)
    cov_right = np.cov(px_right.T)

    cov_top = np.cov(px_top.T)
    cov_bottom = np.cov(px_bottom.T)

    cov_left = np.linalg.inv(cov_left)
    cov_right = np.linalg.inv(cov_right)

    cov_top = np.linalg.inv(cov_top)
    cov_bottom = np.linalg.inv(cov_bottom)

    u_left = np.zeros(sal.shape)
    u_right = np.zeros(sal.shape)
    u_top = np.zeros(sal.shape)
    u_bottom = np.zeros(sal.shape)

    u_final = np.zeros(sal.shape)
    img_lab_unrolled = img_lab.reshape(img_lab.shape[0] * img_lab.shape[1], 3)

    px_mean_left_2 = np.zeros((1, 3))
    px_mean_left_2[0, :] = px_mean_left

    u_left = scipy.spatial.distance.cdist(
        img_lab_unrolled, px_mean_left_2, 'mahalanobis', VI=cov_left)
    u_left = u_left.reshape((img_lab.shape[0], img_lab.shape[1]))

    px_mean_right_2 = np.zeros((1, 3))
    px_mean_right_2[0, :] = px_mean_right

    u_right = scipy.spatial.distance.cdist(
        img_lab_unrolled, px_mean_right_2, 'mahalanobis', VI=cov_right)
    u_right = u_right.reshape((img_lab.shape[0], img_lab.shape[1]))

    px_mean_top_2 = np.zeros((1, 3))
    px_mean_top_2[0, :] = px_mean_top

    u_top = scipy.spatial.distance.cdist(
        img_lab_unrolled, px_mean_top_2, 'mahalanobis', VI=cov_top)
    u_top = u_top.reshape((img_lab.shape[0], img_lab.shape[1]))

    px_mean_bottom_2 = np.zeros((1, 3))
    px_mean_bottom_2[0, :] = px_mean_bottom

    u_bottom = scipy.spatial.distance.cdist(
        img_lab_unrolled, px_mean_bottom_2, 'mahalanobis', VI=cov_bottom)
    u_bottom = u_bottom.reshape((img_lab.shape[0], img_lab.shape[1]))

    max_u_left = np.max(u_left)
    max_u_right = np.max(u_right)
    max_u_top = np.max(u_top)
    max_u_bottom = np.max(u_bottom)

    u_left = u_left / max_u_left
    u_right = u_right / max_u_right
    u_top = u_top / max_u_top
    u_bottom = u_bottom / max_u_bottom

    u_max = np.maximum(
        np.maximum(np.maximum(u_left, u_right), u_top), u_bottom)

    u_final = (u_left + u_right + u_top + u_bottom) - u_max

    u_max_final = np.max(u_final)
    sal_max = np.max(sal)
    sal = sal / sal_max + u_final / u_max_final

  #postprocessing
  # apply centredness map
  sal = sal / np.max(sal)
  s = np.mean(sal)
  alpha = 50.0
  delta = alpha * math.sqrt(s)
  xv, yv = np.meshgrid(np.arange(sal.shape[1]), np.arange(sal.shape[0]))
  (w, h) = sal.shape
  w2 = w / 2.0
  h2 = h / 2.0
  C = 1 - np.sqrt(np.power(xv - h2, 2) + np.power(yv - w2, 2)) / math.sqrt(
      np.power(w2, 2) + np.power(h2, 2))
  sal = sal * C

  #increase bg/fg contrast 
  def f(x):
    b = 10.0
    return 1.0 / (1.0 + math.exp(-b * (x - 0.5)))

  fv = np.vectorize(f)
  sal = sal / np.max(sal)
  sal = fv(sal)
  return sal * 255.0



"""Matcher that handles large scale search.
"""

from owl.features import common as feat_common
from owl.search import common as search_common
from owl.search import simple_matchers


class LargeScaleMatcher(common.BaseMatcher):
  """Matchers that work with large scale features.

  1) 
  """
  params = None

  def __init__(self, matcher_params=None):
    self.params = matcher_params

  


"""Tumblr api.

Wrapper above offical api.
Install: pip install git+https://github.com/tumblr/pytumblr
"""

import sys

import pytumblr

from owl.data import common


class TumblrAPI(object):
  """High-level task manager using official tumblr api.

  NOTE: tumblr limites 1000 requests per hour and 5000 per day.
  """
  # keys: associated with eyestyle app.
  consumer_key = "rXnjXHodcwWevFSyn4SPna57z11Si68lhDXLEPALaIVxBtVrwe"
  consumer_secret = "lpANgcmY0KtWJjESMNTbeJpHCJawDz1LpHkLcLTLSnmcQJjO3k"
  token = "lLImTDSPUkwPUK63fySRv0NYHJBmF7NlMnAMV9g7IB8NxbQH4y"
  token_secret = "9xT1CMdO6rrxt11rI6ydQVRi8G15WB0MNgWBaVnGrjfblJQTJH"
  api_key = "rXnjXHodcwWevFSyn4SPna57z11Si68lhDXLEPALaIVxBtVrwe"
  # client object.
  client = None
  # key used to continue crawl.
  cont_key = None

  def __init__(self):
    # init client.
    self.client = pytumblr.TumblrRestClient(
        self.consumer_key, self.consumer_secret, self.token, self.token_secret)

  def set_initial_cont_key(self, key):
    """Set the initial continue key to start.
    """
    self.cont_key = key

  def crawl_posts(self, tag, only_photo=True, to_continue=False):
    """Crawl posts given a specific tag.

    Args:
      tag: keyword used to crawl images.
      num: number of posts to return.
      only_photo: only return photo post.
      cont_key: key used to continue fetch.
    Returns:
      items meet options.
    """
    try:
      # make request.
      if to_continue and self.cont_key is not None:
        res = self.client.tagged(tag, before=self.cont_key)
      else:
        res = self.client.tagged(tag)
      # exceeds limit
      if "meta" in res:
        raise ValueError("exceeds tumblr limit")
      # parse objects.
      items = []
      oldest_timestamp = sys.maxint
      for item in res:
        if only_photo and item["type"] != "photo":
          continue
        if "featured_timestamp" in item:
          oldest_timestamp = min(item["featured_timestamp"], oldest_timestamp)
        else:
          oldest_timestamp = min(item["timestamp"], oldest_timestamp)
        # parse image urls.
        if "photos" in item:
          img_urls = []
          for photo in item["photos"]:
            img_urls.append(photo["original_size"]["url"])
          item["img_urls"] = img_urls
        items.append(item)
      if to_continue:
        self.cont_key = oldest_timestamp
      return items
    except Exception as ex:
      raise Exception(
          common.get_detailed_error_msg(ex.message, sys.exc_info()))



"""Implementation of "Frequency-tuned Salient Region Detection",
by R. Achanta, S. Hemami, F. Estrada and S. Susstrunk.
Code from https://github.com/yhenon/pyimgsaliency.
"""

import numpy as np
import skimage
import skimage.io
from skimage.util import img_as_float
import scipy.signal


def get_saliency_ft(img_arr):
  """compute saliency with ft.
  """
  img = img_arr
  img_rgb = img_as_float(img)
  img_lab = skimage.color.rgb2lab(img_rgb)
  mean_val = np.mean(img_rgb, axis=(0, 1))

  kernel_h = (1.0 / 16.0) * np.array([[1, 4, 6, 4, 1]])
  kernel_w = kernel_h.transpose()
  blurred_l = scipy.signal.convolve2d(img_lab[:, :, 0], kernel_h, mode='same')
  blurred_a = scipy.signal.convolve2d(img_lab[:, :, 1], kernel_h, mode='same')
  blurred_b = scipy.signal.convolve2d(img_lab[:, :, 2], kernel_h, mode='same')
  blurred_l = scipy.signal.convolve2d(blurred_l, kernel_w, mode='same')
  blurred_a = scipy.signal.convolve2d(blurred_a, kernel_w, mode='same')
  blurred_b = scipy.signal.convolve2d(blurred_b, kernel_w, mode='same')

  im_blurred = np.dstack([blurred_l, blurred_a, blurred_b])
  sal = np.linalg.norm(mean_val - im_blurred, axis=2)
  sal_max = np.max(sal)
  sal_min = np.min(sal)
  sal = 255 * ((sal - sal_min) / (sal_max - sal_min))
  return sal



"""Image processing code.

NOTE: all base64 string is encoded from binary data.
"""

import base64
import hashlib
import mimetypes
import PIL.Image

import io
from cStringIO import StringIO
import urllib2

import numpy as np
""" IO """


def read_img_bin(img_fn):
  """Get image binary from an image file.

  Args:
    img_fn: image file path.
  Returns:
    image binary data.
  """
  with open(img_fn, "rb") as f:
    img_bin_data = f.read()
    return img_bin_data


def read_img_arr(img_fn):
  """Read image file to get ndarray.
  """
  img_bin = read_img_bin(img_fn)
  img_arr = img_bin_to_img_arr(img_bin)
  return img_arr


def read_img_base64(img_fn):
  """Read image file to get base64 string.
  """
  img_bin = read_img_bin(img_fn)
  img_base64 = img_bin_to_base64(img_bin)
  return img_base64


def write_img_arr(img_arr, save_fn):
  """Write image array to file.

  Args:
    img_arr: numpy array of image.
    save_fn: file to save image.
  """
  pil_img = img_arr_to_pil_img(img_arr)
  pil_img.save(save_fn)


def download_img_from_url(img_url):
  """Download image data from url.

  Args:
    img_url: url of the image.
  Returns:
    binary string of image, image format.
  """
  user_agent = "Mozilla/5.0 (Windows; U; Windows NT 5.1; en-US; rv:1.9.0.7) Gecko/2009021910 Firefox/3.0.7"
  header = {'User-agent': user_agent}
  res_data = urllib2.Request(img_url, None, header)
  res_fh = urllib2.urlopen(res_data)
  img_bin = res_fh.read()
  content_type = res_fh.headers["content-type"]
  mimetypes.init()
  img_ext = mimetypes.guess_extension(content_type)
  # remove dot prefix.
  img_ext = img_ext[1:]
  if img_ext == "jpe":
    img_ext = "jpg"
  return img_bin, img_ext


""" Type convertion """


def img_bin_to_base64(img_bin):
  """Convert image binary data to base64.

  Args:
    img_bin: binary image data/string.
  Returns:
    base64 of image data.
  """
  img_base64 = base64.b64encode(img_bin)
  return img_base64


def img_bin_to_img_arr(img_bin, use_grayscale=False):
  """Convert image binary data to numpy array.

  Args:
    img_bin: binary image data.
    use_grayscale: convert to grayscale.
  Retunrs:
    numpy array: (height, width, chs).
  """
  pil_img = PIL.Image.open(StringIO(img_bin))
  if use_grayscale:
    new_img = pil_img.convert("L")
  else:
    new_img = pil_img.convert("RGB")
  return np.asarray(new_img)


def img_bin_to_sha1(img_bin):
  """Hash image binary data to sha1.

  Args:
    img_bin: binary string of image.
  Returns:
    sha1 of the image.
  """
  img_sha1 = hashlib.sha1(img_bin).hexdigest()
  return img_sha1


def base64_to_img_bin(img_base64):
  """Decode base64 image to binary string.

  Args:
    img_base64: base64 image string.
  Returns:
    binary image data.
  """
  img_bin = base64.b64decode(img_base64)
  # or: img_bin = img_base64.decode("base64")
  return img_bin


def base64_to_sha1(img_base64):
  """Hash base64 image to sha1.

  Args:
    img_base64: base64 string of image.
  Returns:
    sha1 of the image.
  """
  img_sha1 = hashlib.sha1(img_base64).hexdigest()
  return img_sha1


def base64_to_img_arr(img_base64, use_grayscale=False):
  """Convert base64 image to rgb numpy array.

  Args:
    img_base64: base64 image string.
    use_grayscale: convert to grayscale.
  Returns:
    numpy array: (height, width, chs)
  """
  img_bin_str = base64_to_img_bin(img_base64)
  return img_bin_to_img_arr(img_bin_str, use_grayscale)


def base64_to_data_uri(img_base64):
  """For display in html.
  """
  datauri = "data:image/jpg;base64," + img_base64
  return datauri


def img_arr_to_base64(img_arr):
  """Convert numpy array image to base64.
  """
  # arr to bin
  pil_img = img_arr_to_pil_img(img_arr)
  img_bin = pil_img_to_img_bin(pil_img)
  # bin to base64
  img_base64 = img_bin_to_base64(img_bin)
  return img_base64


def img_arr_to_pil_img(img_arr):
  """Convert numpy array image to pil image.
  """
  pil_img = PIL.Image.fromarray(img_arr)
  return pil_img


def pil_img_to_img_arr(pil_img, use_grayscale=False):
  """Convert pil image to numpy array.
  """
  if use_grayscale:
    new_img = pil_img.convert("L")
  else:
    new_img = pil_img.convert("RGB")
  return np.asarray(new_img)


def pil_img_to_img_bin(pil_img):
  """Convert pil image to binary string.
  """
  output = io.BytesIO()
  pil_img.save(output, format="JPEG")
  img_bin = output.getvalue()
  return img_bin


def show_img_arr(img_arr, title):
  """Display image array.

  Args:
    img_arr: numpy array of image.
    title: title name of the display window.
  """
  pil_img = PIL.Image.fromarray(img_arr)
  pil_img.show(title=title)


""" data processing. """


def get_new_dim(imgw, imgh, max_dim=400):
  """Get new image dimension with maximum constraint.

  Args:
    imgw: image width.
    imgh: image height.
    max_dim: maximum image dimension after.
  Returns:
    new img width and height.
  """
  new_imgw = imgw
  new_imgh = imgh
  if imgw > imgh:
    new_imgw = max_dim
    new_imgh = imgh * max_dim / imgw
  if imgh > imgw:
    new_imgh = max_dim
    new_imgw = imgw * max_dim / imgh
  return new_imgw, new_imgh



"""Mechanism to fuse search results from multiple feature index.
"""


class MultiIndexFusion(object):
  """Class to manage combination of multiple result sets.
  """

  def group_and_rerank(self, res_list):
    """Group hits and rerank using distance.

    Args:
      res_list: a list of NNs from each index.
      Each member is: {"data": [(item_id, dist)], "weight": weight_val}
    Returns:
      combined ranking list of items.
    """
    # a predefined value for maximum feature distance.
    max_dist_val = 1000
    num_feats = len(res_list)
    # get feature weights.
    feat_weights = []
    # form item distance dict.
    all_item_dists = {}
    all_item_hits = {}
    for idx, feat_res in enumerate(res_list):
      feat_weights.append(feat_res["weight"])
      for item in feat_res["data"]:
        item_id, item_dist = item
        if item_id not in all_item_hits:
          all_item_hits[item_id] = 0
          all_item_dists[item_id] = [max_dist_val] * num_feats
        # add hits.
        all_item_hits[item_id] += 1
        all_item_dists[item_id][idx] = item_dist
    # normalize weights.
    weight_sum = sum(feat_weights)
    feat_weights = [x / weight_sum for x in feat_weights]
    # group items based on hits and compute weighted dist.
    hit_items = [None] * (num_feats + 1)
    for item_id, hit_num in all_item_hits.iteritems():
      if hit_items[hit_num] is None:
        hit_items[hit_num] = []
      # compute weighted sum distance.
      cur_item_dists = all_item_dists[item_id]
      final_dist = sum(
          [dist * w for dist, w in zip(cur_item_dists, feat_weights)])
      hit_items[hit_num].append((item_id, hit_num, final_dist))
    # rerank each group and form final results.
    final_res = []
    for idx, items in enumerate(hit_items):
      if items is None:
        continue
      hit_items[idx] = sorted(items, key=lambda item: item[2])
      items.reverse()
      # add to final res.
      final_res.extend(items)
    # correct rank from small distance to large.
    final_res.reverse()
    return final_res



"""CNN feature extractor from tensorflow models.
"""

from owl.features import common

from deepmodels.tf.core import commons
from deepmodels.tf.core.dm_models import dm_model_factory
from deepmodels.tf.core.learners import dm_classifier


class CNNFeatParams(common.BaseFeatParams):
  """Parameters for CNN features.
  """
  pass


class CNNFeatExtractor(common.BaseFeatExtractor):
  """CNN feature extractor.
  """

  network = None
  network_params = None

  def start(self):
    """Prepare deep models.
    """
    network = dm_model_factory.get_dm_model(commons.ModelType.INCEPTION_V4)
    learner = dm_classifier.DMClassifier()



"""Python wrapper for Instagram API.

Instagram rate limit:
sandbox: 500 / hour = 1 / 7.2s
live: 5000 / hour

If in sandbox mode, make sure wait for 8s to make the next request.
"""

import requests

import owl.data.common as common


class MediaInfo(object):
  """Data structure holds common media information.

  Attributes:
    likes: number of likes.
    tags: a list of tag names.
    media_type: image, video.
    filter_type: normal.
    img_*: {"url", "width", "height"},
    loc: {"latitude", "longitude", "name"}
  """
  likes = 0
  tags = None
  media_type = None
  filter_type = None
  created_time = None
  caption_text = None
  media_id = None
  img_low_res = None
  img_thumbnail = None
  img_standard = None
  comments = None
  loc = None

  def parse_json(self, raw_json):
    """Parse raw json data to extract attributes.

    Args:
      raw_json: original json returned by the api.
    """
    try:
      self.likes = raw_json["likes"]["count"]
      self.tags = raw_json["tags"]
      self.media_type = raw_json["type"]
      self.filter_type = raw_json["filter"]
      self.created_time = raw_json["created_time"]
      self.caption_text = raw_json["caption"]["text"]
      self.media_id = raw_json["id"]
      self.img_low_res = {
          "url": raw_json["images"]["low_resolution"]["url"],
          "width": raw_json["images"]["low_resolution"]["width"],
          "height": raw_json["images"]["low_resolution"]["height"]
      }
      self.img_thumbnail = {
          "url": raw_json["images"]["thumbnail"]["url"],
          "width": raw_json["images"]["thumbnail"]["width"],
          "height": raw_json["images"]["thumbnail"]["height"]
      }
      self.img_standard = {
          "url": raw_json["images"]["standard_resolution"]["url"],
          "width": raw_json["images"]["standard_resolution"]["width"],
          "height": raw_json["images"]["standard_resolution"]["height"]
      }
      if raw_json["location"] is not None:
        self.loc = {
            "lat": raw_json["location"]["latitude"],
            "long": raw_json["location"]["longitude"],
            "name": raw_json["location"]["name"]
        }
    except Exception as ex:
      print "error parsing media json: {}".format(ex.message)
      print "raw json: {}".format(raw_json)
      raise Exception


class UserInfo(object):
  """Data structure holds user data.
  """
  pass


class Users(object):
  """User endpoints.

  ref url: https://www.instagram.com/developer/endpoints/users/.
  """
  token = None

  def __init__(self, access_token):
    self.token = access_token

  def get_self_info(self):
    """Get information about the owner of the access_token.

    Returns:
      user data in json.
    """
    response = requests.get(
        "https://api.instagram.com/v1/users/self/?access_token={}".format(
            self.token))
    if response.status_code == 200:
      return response.json()["data"]
    else:
      return None

  def get_user_info(self, user_id):
    """Get information about a user.

    Args:
      user_id: user id.
    Returns:
      user data in json.
    """
    response = requests.get(
        "https://api.instagram.com/v1/users/{}/?access_token={}".format(
            user_id, self.token))
    if response.status_code == 200:
      return response.json()["data"]
    else:
      return None

  def get_self_recent_media(self, min_id=None, max_id=None, count=None):
    """Get the most recent media published by the owner of the access_token.

    Args:
      min_id: return media later than this min_id.
      max_id: return media earlier than this max_id.
      count: count of media to return.
    Returns:
      return media information in json.
    """
    url = "https://api.instagram.com/v1/users/self/media/recent/?"
    url = common.append_url_params(url, {
        "min_id": min_id,
        "max_id": max_id,
        "count": count,
        "access_token": self.token
    })
    response = requests.get(url)
    if response.status_code == 200:
      media_infos = []
      for media_json in response.json()["data"]:
        media_info = MediaInfo()
        media_info.parse_json(media_json)
        media_infos.append(media_info)
      return media_infos
    else:
      return None

  def get_user_recent_media(self,
                            user_id,
                            min_id=None,
                            max_id=None,
                            count=None):
    """Get the most recent media published by a user.

    Args:
      user_id: target user.
      min_id: return media later than this min_id.
      max_id: return media earlier than this max_id.
      count: count of media to return.
    Returns:
      return media information in json.
    """
    url = "https://api.instagram.com/v1/users/{}/media/recent/?".format(
        user_id)
    url = common.append_url_params(url, {
        "min_id": min_id,
        "max_id": max_id,
        "count": count,
        "access_token": self.token
    })
    response = requests.get(url)
    if response.status_code == 200:
      media_infos = []
      for media_json in response.json()["data"]:
        media_info = MediaInfo()
        media_info.parse_json(media_json)
        media_infos.append(media_info)
      return media_infos
    else:
      return None

  def get_self_liked_media(self, count=None, max_like_id=None):
    """Get the list of recent media liked by the owner of the access_token.

    Args:
      count: count of media to return.
      max_like_id: return media liked before this id.
    Returns:
      return media information in json.
    """
    url = "https://api.instagram.com/v1/users/self/media/liked?"
    url = common.append_url_params(url, {
        "count": count,
        "max_lik_id": max_like_id,
        "access_token": self.token
    })
    response = requests.get(url)
    if response.status_code == 200:
      media_infos = []
      for media_json in response.json()["data"]:
        media_info = MediaInfo()
        media_info.parse_json(media_json)
        media_infos.append(media_info)
      return media_infos
    else:
      return None

  def search_users(self, query, count=None):
    """Get a list of users matching the query.

    Args:
      query: user query.
      count: number of users to return.
    Returns:
      a list of users in json.
    """
    url = "https://api.instagram.com/v1/users/search?"
    url = common.append_url_params(
        url, {"q": query,
              "count": count,
              "access_token": self.token})
    response = requests.get(url)
    if response.status_code == 200:
      return response.json()["data"]
    else:
      return None



"""Simple object segmentation.
"""

import numpy as np

import cv2


class SimpleObjectSegmentor(object):
  """Class for simple segmentor.

  Segmentation for simple background.
  """

  def boundary_replicate_cut(self, img_rgb):
    """Apply graphcut on image with replicated boundary.

    Args:
      img_rgb: numpy array of rgb image.
    Returns:
      binary mask (uint8) with foreground set to 255.
    """
    # duplicate boundary pixel.
    img_bgr = cv2.cvtColor(img_rgb, cv2.COLOR_RGB2BGR)
    imgh, imgw, _ = img_bgr.shape
    padding = 20
    new_img = cv2.copyMakeBorder(img_bgr, padding, padding, padding, padding,
                                 cv2.BORDER_REPLICATE)
    # run graph cut.
    fg_mask = np.zeros((imgh + 2 * padding, imgw + 2 * padding), np.uint8)
    bgdmodel = np.zeros((1, 65), np.float64)
    fgdmodel = np.zeros((1, 65), np.float64)
    rect = (padding, padding, imgw, imgh)
    cv2.grabCut(new_img, fg_mask, rect, bgdmodel, fgdmodel, 3,
                cv2.GC_INIT_WITH_RECT)
    res_mask = cv2.bitwise_and(fg_mask, 1).astype("uint8")
    res_mask = res_mask[rect[1]:rect[3] + rect[1], rect[0]:rect[2] + rect[0]]
    return res_mask * 255

  def boundary_adapt_replicate_cut(self, img_rgb):
    """Replicate boundary for cut with most frequent pixel.

    Args:
      img_rgb: rgb image ndarray.
    Returns:
      binary mask.
    """
    img_bgr = cv2.cvtColor(img_rgb, cv2.COLOR_RGB2BGR)
    imgh, imgw, _ = img_bgr.shape
    # use most frequent boundary pixel.
    top_row = img_bgr[0, :, :]
    bottom_row = img_bgr[-1, :, :]
    left_col = img_bgr[:, 0, :]
    right_col = img_bgr[:, -1, :]
    boundary_pixels = np.vstack((top_row, bottom_row, left_col, right_col))
    boundary_pixels = boundary_pixels.astype(np.float32)
    term_crit = (cv2.TERM_CRITERIA_EPS, 30, 0.1)
    K = 5
    _, labels, centers = cv2.kmeans(boundary_pixels, K, term_crit, 5,
                                    cv2.KMEANS_PP_CENTERS)
    label_cnt = np.zeros((K), np.int32)
    for label in labels:
      label_cnt[label] += 1
    max_color = centers[np.argmax(label_cnt), :]
    max_color = [np.uint8(v) for v in max_color]
    padding = 20
    dst2 = np.zeros((imgh + 2 * padding, imgw + 2 * padding, 3), np.uint8)
    dst2[:] = max_color
    dst2[padding:padding + imgh, padding:padding + imgw, :] = img_bgr
    # perform cut.
    fg_mask = np.zeros((imgh + 2 * padding, imgw + 2 * padding), np.uint8)
    bgdmodel = np.zeros((1, 65), np.float64)
    fgdmodel = np.zeros((1, 65), np.float64)
    rect = (padding, padding, imgw, imgh)
    cv2.grabCut(dst2, fg_mask, rect, bgdmodel, fgdmodel, 3,
                cv2.GC_INIT_WITH_RECT)
    res_mask = cv2.bitwise_and(fg_mask, 1).astype("uint8")
    res_mask = res_mask[rect[1]:rect[3] + rect[1], rect[0]:rect[2] + rect[0]]
    return res_mask * 255

  def otsu(self, img_rgb):
    """Use OTSU method to do graph cut.
    """
    img_bgr = cv2.cvtColor(img_rgb, cv2.COLOR_RGB2BGR)
    imgh, imgw, _ = img_bgr.shape
    padding = 20
    dst = cv2.copyMakeBorder(img_bgr, padding, padding, padding, padding,
                             cv2.BORDER_REPLICATE)
    dst2_gray = cv2.cvtColor(dst, cv2.COLOR_BGR2GRAY)
    _, thresh = cv2.threshold(dst2_gray, 0, 255,
                              cv2.THRESH_BINARY_INV + cv2.THRESH_OTSU)
    kernel = np.ones((7, 7), np.uint8)
    closed_thresh = cv2.morphologyEx(
        thresh, cv2.MORPH_CLOSE, kernel, iterations=2)
    # cv2.imshow('closed_thresh', closed_thresh)
    # cv2.waitKey(wait_time)
    fg_mask = np.zeros((imgh + padding * 2, imgw + padding * 2, 1), np.uint8)
    fg_mask.fill(cv2.GC_BGD)
    fg_mask.ravel()[closed_thresh.ravel() == 255] = cv2.GC_PR_FGD
    # should we add PR_BGD?
    rect = (padding, padding, imgw, imgh)
    bgdmodel = np.zeros((1, 65), np.float64)
    fgdmodel = np.zeros((1, 65), np.float64)
    cv2.grabCut(dst, fg_mask, rect, bgdmodel, fgdmodel, 2,
                cv2.GC_INIT_WITH_MASK)
    res_mask = cv2.bitwise_and(fg_mask, 1).astype("uint8")
    res_mask = res_mask[rect[1]:rect[3] + rect[1], rect[0]:rect[2] + rect[0]]
    return res_mask * 255



"""General interface for aws.
"""

from owl.third.aws import common as aws_common
from owl.third.aws import s3
from owl.third.aws import dynamodb


class AWSIO(object):
  """Class to talk with aws database for owl.
  """

  session = None
  s3_man = None
  dynamodb = None
  access_key = ""
  secret_key = ""

  def __init__(self, access_key, secret_key, region):
    """Init aws services.
    """
    self.session = aws_common.init_aws(access_key, secret_key, region)
    self.s3_man = s3.S3IO(self.session)
    self.dynamodb_man = dynamodb.DynamoDBIO(self.session)



"""Tests for image tools module.
"""

import unittest

from owl.data import img_tools


class ImgToolsTester(unittest.TestCase):
  def test_img_bin_to_numpy(self):
    img_fn = "/mnt/Lab/imgs/1_25_25389.jpg"
    img_bin = img_tools.read_img_bin(img_fn)
    img_arr = img_tools.img_bin_to_numpy_rgb(img_bin)
    print img_arr.shape
    self.assertLess(img_arr.shape[0], img_arr.shape[1])


if __name__ == "__main__":
  unittest.main()



''' cnn features for retrieval
    jiefeng@2016-02-12
'''

import sys
import numpy as np
import time

from common import *

lib_root = 'E:/Projects/Github/deeplearning/Theano/Lasagne/Scripts/'
print 'deep learning folder: ' + lib_root
sys.path.append(lib_root)

from deep_clf import *


class CNNParams(FeatParamsBase):
  feat_type = FeatType.CNN_VGGS
  feat_name = 'cnn_vggs'


class DeepCNNFeature:

  params = CNNParams()

  def __init__(self):
    pass

  def Init(self, cnn_params):
    self.params = cnn_params
    self.config = ExptConfigBase()
    self.config.model_params.img_sz = (224, 224)
    self.config.model_params.output_layer_name = 'output'
    if self.params.feat_type == FeatType.CNN_VGGS or self.params.feat_type == FeatType.CNN_VGGS_SHORT:
      self.model = DeepClfVGGS(self.config.model_params,
                               self.config.train_params)
    if self.params.feat_type == FeatType.CNN_VGG16 or self.params.feat_type == FeatType.CNN_VGG16_SHORT:
      self.model = DeepClfVGG16(self.config.model_params,
                                self.config.train_params)

  def Start(self):
    self.model.build_model()

  def SetInput(self, img_bgr):
    self.img_bgr = img_bgr

  # compute for image
  def Compute(self, mask=None):
    if mask is not None:
      # get bounding box
      locs = cv2.findNonZero(mask)
      box = cv2.boundingRect(locs)
      self.img_bgr = self.img_bgr[box[0]:box[0] + box[2], box[1]:box[1] + box[
          3]]
    self.model.model_params.output_layer_name = 'fc6'
    cnn_input = self.model.prepare_imgs_for_input(np.array([self.img_bgr]))
    feat = self.model.get_outputs(cnn_input)
    feats = np.absolute(feats)
    return feat

  # compute for a set of images
  def ComputeForFns(self, img_fns, mask_fns=None):
    self.model.model_params.output_layer_name = 'fc6'
    cnn_inputs = self.model.prepare_imgfns_for_input(img_fns)
    feats = self.model.get_outputs(cnn_inputs, 10)
    # have to be nonnegative
    feats = np.absolute(feats)
    feats = sklearn.preprocessing.normalize(
        feats, norm='l2', axis=1, copy=False)
    return feats

  def ComputeConvFeats(self, img_fns):
    # conv5_3
    self.model.model_params.output_layer_name = 'conv5'
    cnn_inputs = self.model.prepare_imgfns_for_input(img_fns)
    feats = self.model.get_outputs(cnn_inputs, 10, False)
    # aggregate conv features
    feats = np.sum(feats, axis=(2, 3))
    print feats.shape
    feats = sklearn.preprocessing.normalize(
        feats, norm='l2', axis=1, copy=False)
    return feats



"""Color histogram.
"""

import numpy as np
import cv2

from owl.features import common


class ColorHistParams(common.BaseFeatParams):
  feat_type = common.FeatType.COLOR_HIST_LINEAR
  bin_nums = [16, 16, 16]
  color_space = common.ColorSpace.RGB


class ColorHistogram(common.BaseFeatExtractor):
  """Color histogram descriptors.
  """
  # separate color range for each channel.
  ch_color_ranges = []
  # concatenated color range.
  concat_color_range = []

  def __init__(self, feat_params):
    self.feat_name = "color_histogram"
    self.feat_type = feat_params.feat_type
    self.feat_params_ = feat_params

  def set_input(self, rgb_img):
    if self.feat_params_.color_space == common.ColorSpace.RGB:
      self.internal_img = rgb_img
      self.ch_color_ranges = [[0, 256], [0, 256], [0, 256]]
      self.concat_color_range = [0, 256, 0, 256, 0, 256]
    if self.feat_params_.color_space == common.ColorSpace.HSV:
      new_img = np.float32(rgb_img)
      new_img /= 255
      self.internal_img = cv2.cvtColor(new_img, cv2.COLOR_RGB2HSV)
      self.ch_color_ranges = [[0, 360], [0, 1], [0, 1]]
      self.concat_color_range = [0, 360, 0, 1, 0, 1]
    if self.feat_params_.color_space == common.ColorSpace.LAB:
      new_img = np.float32(rgb_img)
      new_img /= 255
      self.internal_img = cv2.cvtColor(new_img, cv2.COLOR_RGB2LAB)
      self.ch_color_ranges = [[0, 100], [-127, 127], [-127, 127]]
      self.concat_color_range = [0, 100, -127, 127, -127, 127]

  def compute(self, rgb_img=None, mask=None):
    if rgb_img is not None:
      self.set_input(rgb_img)
    num_pixels = self.internal_img.size / 3
    if mask is not None:
      rgb_img = cv2.resize(rgb_img, (mask.shape[1], mask.shape[0]))
    feat = None
    if self.feat_type == common.FeatType.COLOR_HIST_LINEAR:
      feat = None
      for ch_id in range(3):
        hist = cv2.calcHist([self.internal_img], [ch_id], mask,
                            [self.feat_params_.bin_nums[ch_id]],
                            self.ch_color_ranges[ch_id])
        hist = np.float32(hist) / num_pixels
        if feat is None:
          feat = hist
        else:
          feat = np.append(feat, hist, axis=0)
      feat = feat.flatten()
    if self.feat_type == common.FeatType.COLOR_HIST_JOINT:
      hist = cv2.calcHist([self.internal_img], [0, 1, 2], mask,
                          self.feat_params_.bin_nums, self.concat_color_range)
      hist = np.float32(hist) / num_pixels
      feat = hist.flatten()
    return feat

  def calc_dist(self, feat1, feat2):
    return np.linalg.norm(feat1 - feat2)

  def get_feat_dim(self):
    if self.feat_params_.feat_type == common.FeatType.COLOR_HIST_LINEAR:
      return sum(self.feat_params_.bin_nums)
    else:
      dim = 1
      for bin_num in self.feat_params_.bin_nums:
        dim *= bin_num
      return dim



"""Instagram api test.
"""

import unittest

from owl.third import instagram

access_token = "290412779.f188e2f.d6e2d8c4cde74eab8a96564a3b1fbb14"


class InstagramUserTest(unittest.TestCase):
  """Class to test user endpoints.
  """
  user_api = instagram.Users(access_token)

  def test_get_self(self):
    self_info = self.user_api.get_self_info()
    self.assertIsNotNone(self_info)

  def test_get_self_media(self):
    count = 30
    media_infos = self.user_api.get_self_recent_media(count=count)
    self.assertEqual(len(media_infos), count)


if __name__ == "__main__":
  unittest.main()



"""Dimension reduction processors.
"""

import abc

import os
import cPickle as pickle

from sklearn import decomposition


class DimRedProtocol(object):
  """Ways to do dimension reduction for feature.
  """
  BINARY = 0
  PCA = 1
  INCREMENTAL_PCA = 2
  TSNE = 3


class BaseDimReducer(object):
  """Base class for dim reducer.
  """
  __metaclass__ = abc.ABCMeta

  res_dir = ""
  model_filepath = ""
  model_name = ""
  protocol = None
  model = None

  @abc.abstractmethod
  def __init__(self, res_dir):
    """Init function.

    Override to pass specific parameters.
    Set default model_fn to save.

    Args:
      res_dir: directory to save result data.
    """
    self.res_dir = res_dir
    self.model_filepath = os.path.join(self.res_dir,
                                       "{}.pkl".format(self.model_name))

  @abc.abstractmethod
  def train(self, data, get_transformed=False):
    """Perform reduction.

    Args:
      data: row major numpy array.
      save_fn: relative filename to save the model under res_dir.
      get_transform: get transformed data.
    Returns:
      transformed data if needed.
    """
    pass

  @abc.abstractmethod
  def transform(self, data):
    """Transform input data based on learned reducer.

    Args:
      data: row major data matrix.
    Returns:
      transformed data.
    """
    pass

  def save_load_model(self, model_filepath=None, to_save=True):
    """Manage model save and load from pickle.

    Args:
      model_filepath: target model file to use.
      to_save: whether to save or load model.
    """
    if model_filepath is None:
      model_filepath = self.model_filepath
    if to_save:
      with open(model_filepath, "wb") as f:
        pickle.dump(self.model, f)
        print "model {} saved to {}".format(self.model_name, model_filepath)
    else:
      with open(model_filepath, "rb") as f:
        self.model = pickle.load(f)
        print "model {} loaded from {}".format(self.model_name, model_filepath)


class PCADimReducer(BaseDimReducer):
  """Class for dimensionality reduction.
  """

  def __init__(self, res_dir, num_comp=None):
    self.protocol = DimRedProtocol.PCA
    self.model_name = "PCA"
    self.model = decomposition.PCA(num_comp)
    super(PCADimReducer, self).__init__(res_dir)

  def train(self, data, get_transformed=False):
    """Train PCA model.
    """
    transformed_data = None
    print "fitting pca to data..."
    if get_transformed:
      transformed_data = self.model.fit_transform(data)
    else:
      self.model.fit(data)
    print "pca done."
    # save pca object.
    self.save_load_model(to_save=True)
    return transformed_data

  def transform(self, data):
    """Apply pca to new data.

    Args:
      data: input data to be transfomred, numpy array.
    Returns:
      transformed data.
    """
    if self.model is None and os.path.exists(self.model_filepath):
      self.save_load_model(self.model_filepath, to_save=False)
    return self.model.transform(data)


class IncrementalPCADimReducer(BaseDimReducer):
  """Incremental PCA class.
  """

  def __init__(self, res_dir, num_comp, batch_size=None, whiten=False):
    self.protocol = DimRedProtocol.INCREMENTAL_PCA
    self.model_name = "Incremental_PCA"
    if batch_size is None:
      batch_size = 2 * num_comp
    assert batch_size > num_comp, \
    "batch size must be larger than num components"
    self.model = decomposition.IncrementalPCA(
        n_components=num_comp, batch_size=batch_size, whiten=whiten)
    super(IncrementalPCADimReducer, self).__init__(res_dir)

  def train(self, data, get_transformed=False):
    """Train incremental PCA by feeding batch data.
    """
    transformed_data = None
    print "fitting incremental pca to data..."
    self.model.partial_fit(data)
    print "incremental pca done."
    if get_transformed:
      transformed_data = self.model.transform(data)
    # save model.
    self.save_load_model(to_save=True)
    return transformed_data

  def transform(self, data):
    """Apply incremental pca to new data.

    Args:
      data: input data to be transfomred, numpy array.
    Returns:
      transformed data.
    """
    if self.model is None and os.path.exists(self.model_filepath):
      self.save_load_model(self.model_filepath, to_save=False)
    return self.model.transform(data)



"""Shared services for aws.
"""

import json
import boto3


# TODO(jiefeng): to finish function.
def read_config_fn(config_fn):
  """Read configuration from file.
  """
  with open(config_fn, "r") as f:
    credential = json.load(f)
    return credential


def init_aws(access_key, secret_key, region):
  """Create aws session.
  """
  session = boto3.Session(
      aws_access_key_id=access_key,
      aws_secret_access_key=secret_key,
      region_name=region)
  return session



"""OOP interface for DynamoDB.
"""

from owl.third.aws import common as aws_common


class DynamoDBIO(object):
  """Class for managing dynamodb tasks.
  """
  session = None
  client = None
  table = None

  def __init__(self, session=None, access_key="", secret_key="", region=""):
    """Create dynamodb client.
    """
    if session is not None:
      self.session = session
    else:
      self.session = aws_common.init_aws(access_key, secret_key, region)
    self.client = self.session.resource("dynamodb")

  def use_table(self, table_name):
    """Set the table object.
    """
    self.table = self.client.Table(table_name)
    assert self.table.table_status == "ACTIVE", "dynamodb table {} not exist".format(
        table_name)
    print "using table: {}".format(table_name)

  def get_table_item_count(self):
    """Get how many items are in the current table.

    Updated every 6 hours on aws.
    """
    return self.table.item_count

  def check_table_exist(self, table_name):
    """Check whether table exists.

    Args:
      table_name: name of the target table.
    Returns:
      boolean value indicating existance.
    """
    try:
      self.use_table(table_name)
      return True
    except:
      return False

  def create_table(self,
                   table_name,
                   partition_key,
                   sort_key=None,
                   read_cap=1,
                   write_cap=1):
    """Create a table.
    """
    # check if table already exists.
    if self.check_table_exist(table_name):
      self.use_table(table_name)
    else:
      # create one.
      key_schema = [{"AttributeName": partition_key, "KeyType": "HASH"}]
      attributes = [{"AttributeName": partition_key, "AttributeType": "S"}]
      if sort_key is not None:
        key_schema.append({"AttributeName": sort_key, "KeyType": "RANGE"})
        attributes.append({{"AttributeName": sort_key, "AttributeType": "S"}})
      self.table = self.client.create_table(
          TableName=table_name,
          KeySchema=key_schema,
          # Only need to define keys type
          AttributeDefinitions=attributes,
          ProvisionedThroughput={
              "ReadCapacityUnits": read_cap,
              "WriteCapacityUnits": write_cap
          })
    print "table {} created.".format(table_name)

  def add_item(self, item):
    """Write attributes to current table.

    Args:
      item: name value pairs.
    """
    self.table.put_item(Item=item)

  def add_batch_items(self, items):
    """Write a batch of items to current table.

    Args:
      items: list of item.
    """
    with self.table.batch_writer() as batch:
      for cur_item in items:
        batch.put_item(Item=cur_item)

  def get_item(self, key_dict, attributes_to_get=None):
    """Get attributes from table.

    Args:
      key_dict: dictionary of primary key.
      attributes_to_get: projection expression to specify attributes.
    Returns:
      retrieved item with specified attributes.
    """
    if attributes_to_get is None:
      res = self.table.get_item(Key=key_dict)
    else:
      res = self.table.get_item(
          Key=key_dict, AttributesToGet=attributes_to_get)
    if "Item" in res:
      return res["Item"]
    else:
      return None

  def scan_items(self, cont_key=None):
    """Scan table and return items.

    Args:
      cont_key: key used to continue scan.
    Returns:
      current batch of items, continue_key (None if not exist).
    """
    if cont_key is not None:
      response = self.table.scan(ExclusiveStartKey=cont_key)
    else:
      response = self.table.scan(Limit=10)
    cont_key = None
    if "LastEvaluatedKey" in response:
      cont_key = response["LastEvaluatedKey"]
    return response["Items"], cont_key



"""Implementation of "Saliency Optimization from Robust Background Detection",
by Wangjiang Zhu, Shuang Liang, Yichen Wei and Jian Sun.
code from https://github.com/yhenon/pyimgsaliency.
"""

import math
import networkx as nx
import numpy as np
import scipy.spatial.distance
import scipy.signal
import skimage
import skimage.io
from skimage.segmentation import slic
from skimage.util import img_as_float


def S(x1, x2, geodesic, sigma_clr=10):
  return math.exp(-pow(geodesic[x1, x2], 2) / (2 * sigma_clr * sigma_clr))


def compute_saliency_cost(smoothness, w_bg, wCtr):
  n = len(w_bg)
  A = np.zeros((n, n))
  b = np.zeros((n))

  for x in xrange(0, n):
    A[x, x] = 2 * w_bg[x] + 2 * (wCtr[x])
    b[x] = 2 * wCtr[x]
    for y in xrange(0, n):
      A[x, x] += 2 * smoothness[x, y]
      A[x, y] -= 2 * smoothness[x, y]

  x = np.linalg.solve(A, b)

  return x


def path_length(path, G):
  dist = 0.0
  for i in xrange(1, len(path)):
    dist += G[path[i - 1]][path[i]]['weight']
  return dist


def make_graph(grid):
  # get unique labels
  vertices = np.unique(grid)

  # map unique labels to [1,...,num_labels]
  reverse_dict = dict(zip(vertices, np.arange(len(vertices))))
  grid = np.array([reverse_dict[x] for x in grid.flat]).reshape(grid.shape)

  # create edges
  down = np.c_[grid[:-1, :].ravel(), grid[1:, :].ravel()]
  right = np.c_[grid[:, :-1].ravel(), grid[:, 1:].ravel()]
  all_edges = np.vstack([right, down])
  all_edges = all_edges[all_edges[:, 0] != all_edges[:, 1], :]
  all_edges = np.sort(all_edges, axis=1)
  num_vertices = len(vertices)
  edge_hash = all_edges[:, 0] + num_vertices * all_edges[:, 1]
  # find unique connections
  edges = np.unique(edge_hash)
  # undo hashing
  edges = [[vertices[x % num_vertices], vertices[x / num_vertices]]
           for x in edges]

  return vertices, edges


def get_saliency_rbd(img_arr):
  """Saliency map computation.
  """
  # img = skimage.io.imread(img_path)
  img = img_arr

  if len(img.shape) != 3:  # got a grayscale image
    img = skimage.color.gray2rgb(img)

  img_lab = img_as_float(skimage.color.rgb2lab(img))

  img_rgb = img_as_float(img)

  img_gray = img_as_float(skimage.color.rgb2gray(img))

  segments_slic = slic(
      img_rgb,
      n_segments=250,
      compactness=10,
      sigma=1,
      enforce_connectivity=False)

  num_segments = len(np.unique(segments_slic))

  nrows, ncols = segments_slic.shape
  max_dist = math.sqrt(nrows * nrows + ncols * ncols)

  grid = segments_slic

  (vertices, edges) = make_graph(grid)

  gridx, gridy = np.mgrid[:grid.shape[0], :grid.shape[1]]

  centers = dict()
  colors = dict()
  distances = dict()
  boundary = dict()

  for v in vertices:
    centers[v] = [gridy[grid == v].mean(), gridx[grid == v].mean()]
    colors[v] = np.mean(img_lab[grid == v], axis=0)

    x_pix = gridx[grid == v]
    y_pix = gridy[grid == v]

    if np.any(x_pix == 0) or np.any(y_pix == 0) or np.any(
        x_pix == nrows - 1) or np.any(y_pix == ncols - 1):
      boundary[v] = 1
    else:
      boundary[v] = 0

  G = nx.Graph()

  #buid the graph
  for edge in edges:
    pt1 = edge[0]
    pt2 = edge[1]
    color_distance = scipy.spatial.distance.euclidean(colors[pt1], colors[pt2])
    G.add_edge(pt1, pt2, weight=color_distance)

  #add a new edge in graph if edges are both on boundary
  for v1 in vertices:
    if boundary[v1] == 1:
      for v2 in vertices:
        if boundary[v2] == 1:
          color_distance = scipy.spatial.distance.euclidean(colors[v1],
                                                            colors[v2])
          G.add_edge(v1, v2, weight=color_distance)

  geodesic = np.zeros((len(vertices), len(vertices)), dtype=float)
  spatial = np.zeros((len(vertices), len(vertices)), dtype=float)
  smoothness = np.zeros((len(vertices), len(vertices)), dtype=float)
  adjacency = np.zeros((len(vertices), len(vertices)), dtype=float)

  sigma_clr = 10.0
  sigma_bndcon = 1.0
  sigma_spa = 0.25
  mu = 0.1

  all_shortest_paths_color = nx.shortest_path(
      G, source=None, target=None, weight='weight')

  for v1 in vertices:
    for v2 in vertices:
      if v1 == v2:
        geodesic[v1, v2] = 0
        spatial[v1, v2] = 0
        smoothness[v1, v2] = 0
      else:
        geodesic[v1, v2] = path_length(all_shortest_paths_color[v1][v2], G)
        spatial[v1, v2] = scipy.spatial.distance.euclidean(
            centers[v1], centers[v2]) / max_dist
        smoothness[v1, v2] = math.exp(-(geodesic[v1, v2] * geodesic[v1, v2]) /
                                      (2.0 * sigma_clr * sigma_clr)) + mu

  for edge in edges:
    pt1 = edge[0]
    pt2 = edge[1]
    adjacency[pt1, pt2] = 1
    adjacency[pt2, pt1] = 1

  for v1 in vertices:
    for v2 in vertices:
      smoothness[v1, v2] = adjacency[v1, v2] * smoothness[v1, v2]

  area = dict()
  len_bnd = dict()
  bnd_con = dict()
  w_bg = dict()
  ctr = dict()
  wCtr = dict()

  for v1 in vertices:
    area[v1] = 0
    len_bnd[v1] = 0
    ctr[v1] = 0
    for v2 in vertices:
      d_app = geodesic[v1, v2]
      d_spa = spatial[v1, v2]
      w_spa = math.exp(-((d_spa) * (d_spa)) / (2.0 * sigma_spa * sigma_spa))
      area_i = S(v1, v2, geodesic)
      area[v1] += area_i
      len_bnd[v1] += area_i * boundary[v2]
      ctr[v1] += d_app * w_spa
    bnd_con[v1] = len_bnd[v1] / math.sqrt(area[v1])
    w_bg[v1] = 1.0 - math.exp(-(bnd_con[v1] * bnd_con[v1]) /
                              (2 * sigma_bndcon * sigma_bndcon))

  for v1 in vertices:
    wCtr[v1] = 0
    for v2 in vertices:
      d_app = geodesic[v1, v2]
      d_spa = spatial[v1, v2]
      w_spa = math.exp(-(d_spa * d_spa) / (2.0 * sigma_spa * sigma_spa))
      wCtr[v1] += d_app * w_spa * w_bg[v2]

  # normalise value for wCtr

  min_value = min(wCtr.values())
  max_value = max(wCtr.values())

  minVal = [key for key, value in wCtr.iteritems() if value == min_value]
  maxVal = [key for key, value in wCtr.iteritems() if value == max_value]

  for v in vertices:
    wCtr[v] = (wCtr[v] - min_value) / (max_value - min_value)

  img_disp1 = img_gray.copy()
  img_disp2 = img_gray.copy()

  x = compute_saliency_cost(smoothness, w_bg, wCtr)

  for v in vertices:
    img_disp1[grid == v] = x[v]

  img_disp2 = img_disp1.copy()
  sal = np.zeros((img_disp1.shape[0], img_disp1.shape[1], 3))

  sal = img_disp2
  sal_max = np.max(sal)
  sal_min = np.min(sal)
  sal = 255 * ((sal - sal_min) / (sal_max - sal_min))

  return sal



"""Basic matchers.
"""

import os

import cPickle as pickle
import numpy as np
from scipy.spatial import distance

from owl.search import common

from annoy import AnnoyIndex


class ExactMatcher(common.BaseMatcher):
  """Matchers use common distance metric.
  """

  db_feats = None
  db_metadata = None
  matcher_protocol = common.MatchProtocol.EXACT

  def __init__(self, name, matcher_type, res_dir=None):
    super(ExactMatcher, self).__init__(name, matcher_type, res_dir)

  def prepare_database(self, db_items):
    """Load database.

    Args:
      db_items: a dictionary with keys:
      "metadata", a list of metadata for each item;
      "features": a numpy array of feature vectors, row basis.
    """
    self.db_feats = db_items["features"]
    self.db_metadata = db_items["metadata"]

  def match(self, query_items, top_nn=10, dist_func=None):
    """Do match to get top neighbors.

    Args:
      query_items: dict with "metadata" and "features".
      top_nn: number of neighbors to get.
      dist_func: custom distance function which takes two
      feature vectors and return a value.
    Returns:
      a 2d list where (i,j) element is the jth most similar item for ith query.
    """
    query_feats = query_items["features"]
    # Compute distance matrix.
    dist_mat = None
    if dist_func is None:
      if self.matcher_type == common.MatcherType.L2:
        dist_mat = distance.cdist(query_feats, self.db_feats, "euclidean")
      if self.matcher_type == common.MatcherType.L1:
        dist_mat = distance.cdist(query_feats, self.db_feats, "minkowski", 1)
    else:
      db_size = len(self.db_feats)
      query_size = len(query_feats)
      dist_mat = np.zeros((query_size, db_size), dtype=float)
      for qid in range(query_size):
        for did in range(db_size):
          dist_mat[qid, did] = dist_func(query_feats[qid], self.db_feats[did])
    # sort results.
    ranked_lists = np.argsort(dist_mat, axis=1)
    return ranked_lists[:, :top_nn]


class ANNMatcher(common.BaseMatcher):
  """Matcher using approximate nearest neighbors.
  """

  index_fn = None
  metadata_fn = None
  res_dir = None
  # list of metadata objects for all items.
  db_metadata = None
  # a row major matrix of feature vectors.
  db_feats = None
  db_index = None
  feat_dim = 0
  num_trees = 30
  match_protocol = common.MatchProtocol.APPROXIMATE

  def __init__(self, name, matcher_type, feat_dim, num_trees, res_dir=None):
    super(ANNMatcher, self).__init__(name, matcher_type, res_dir)
    self.index_fn = os.path.join(self.res_dir,
                                 "{}_annoy_ann.index".format(name))
    assert self.matcher_type in [
        common.MatcherType.L2, common.MatcherType.ANGULAR
    ], "current ann matcher only supports l2 or angular distance"
    # create index object.
    assert feat_dim > 0, "feature dimension can not be smaller than 0."
    self.feat_dim = feat_dim
    self.num_trees = num_trees
    if self.matcher_type == common.MatcherType.L2:
      self.db_index = AnnoyIndex(self.feat_dim, "euclidean")
    else:
      self.db_index = AnnoyIndex(self.feat_dim, "angular")
    self.metadata_fn = os.path.join(self.res_dir,
                                    self.matcher_name + ".metadata")

  def load_index(self):
    """Load index from file.
    """
    self.db_index.load(self.index_fn)
    print "db index loaded."
    with open(self.metadata_fn, "rb") as f:
      self.db_metadata = pickle.load(f)
      print "metadata loaded."

  def prepare_database(self, db_items):
    """Create index and save to file.

    Args:
      db_items: a dictionary with keys:
      "metadata", a list of metadata for each item, e.g. objectid.
      "features": a numpy array of feature vectors, row basis.
    """
    if db_items is None and os.path.exists(self.index_fn):
      self.load_index()
    else:
      self.db_metadata = db_items["metadata"]
      self.db_feats = db_items["features"]
      feat_num, feat_dim = self.db_feats.shape
      assert feat_dim == self.feat_dim, \
      "database feature dimension is not the same as index."
      # add items to index.
      print "start building ann index..."
      for i in range(feat_num):
        self.db_index.add_item(i, self.db_feats[i])
      # build index.
      # TODO(jiefeng): choose a proper number of tree.
      self.db_index.build(self.num_trees)
      print "db index built."
      self.db_index.save(self.index_fn)
      print "index file saved to {}".format(self.index_fn)
      with open(self.metadata_fn, "wb") as f:
        pickle.dump(self.db_metadata, f)
        print "metadata saved."

  def match(self, query_items, top_nn=10, dist_func=None):
    """Perform match on database for each query.

    Args:
      query_items: a dict with "metadata" and "features".
    Returns:
      a ranked list of db item ids.
    """
    query_feats = query_items["features"]
    query_size = query_feats.shape[0]
    print "start matching..."
    ranked_list = np.zeros((query_size, top_nn), dtype=int)
    for query_id, query_feat in enumerate(query_feats):
      # cur_nns: ([ids],[dists])
      cur_nns = self.db_index.get_nns_by_vector(
          query_feat, top_nn, include_distances=True)
      nn_ids, _ = cur_nns
      ranked_list[query_id] = np.array(nn_ids, dtype=int)
    print "matching done."
    return ranked_list



from setuptools import setup, find_packages

with open("requirements.txt", "r") as f:
  dep_packages = f.readlines()
  dep_packages = [x.strip() for x in dep_packages]

setup(
    name="projectowl",
    version="0.3.5",
    description="high level computer vision library",
    keywords="computer vision image feature pipeline",
    url="https://github.com/flyfj/owl",
    author="Jie Feng",
    author_email="jiefengdev@gmail.com",
    license="MIT",
    packages=find_packages("./"),
    install_requires=dep_packages,
    include_package_data=True,
    zip_safe=False)



"""Test search components.
"""

import unittest

from owl.search import multi_index_fusion


class MultiIndexFusionTester(unittest.TestCase):
  def test_fusion(self):
    dummy_res = [None] * 3
    dummy_res[0] = {
        "data": [("1", 0.32), ("3", 0.22), ("4", 0.98)],
        "weight": 0.5
    }
    dummy_res[1] = {"data": [("1", 0.45), ("4", 0.34)], "weight": 0.2}
    dummy_res[2] = {"data": [("1", 0.3), ("3", 0.65)], "weight": 0.8}
    fuser = multi_index_fusion.MultiIndexFusion()
    final_res = fuser.group_and_rerank(dummy_res)
    print final_res


if __name__ == "__main__":
  unittest.main()
